@@ -2,6 +2,7 @@
  * ProFTPD: mod_sql_postgres -- Support for connecting to Postgres databases.
  * Time-stamp: <1999-10-04 03:21:21 root>
  * Copyright (c) 2001 Andrew Houghton
+ * Copyright (c) 2004-2013 TJ Saunders
  *  
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -15,36 +16,53 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
  *
  * As a special exemption, Andrew Houghton and other respective copyright
  * holders give permission to link this program with OpenSSL, and distribute
  * the resulting executable, without including the source code for OpenSSL in
  * the source distribution.
  *
- * $Id: mod_sql_postgres.c,v 1.18 2003/07/07 20:42:53 castaglia Exp $
- */
-
-/*
+ * $Id: mod_sql_postgres.c,v 1.56 2013/10/09 04:57:13 castaglia Exp $
  * $Libraries: -lm -lpq $
  */
 
-/* 
- * Internal define used for debug and logging.  All backends are encouraged
+/* Internal define used for debug and logging.  All backends are encouraged
  * to use the same format.
  */
-#define _MOD_VERSION "mod_sql_postgres/4.02"
+#define MOD_SQL_POSTGRES_VERSION	"mod_sql_postgres/4.0.4"
 
 #define _POSTGRES_PORT "5432"
 
-#include <libpq-fe.h>
 #include "conf.h"
 #include "../contrib/mod_sql.h"
 
+#include <libpq-fe.h>
+
+/* For the pg_encoding_to_char() function, used for NLS support, we need
+ * to include the <mb/pg_wchar.h> file.  It's OK; the function has been
+ * declared in that file for a while, according to:
+ *
+ *  http://archives.postgresql.org/pgsql-bugs/2006-07/msg00125.php
+ *
+ * But it's a pain to quell all of the compiler warnings about redefined
+ * this, undefined that.  The linker finds the symbol without issue, so
+ * punt on including <mb/pg_wchar.h> file.  Instead, we'll copy the
+ * function declaration here.
+ */
+#ifdef PR_USE_NLS
+extern const char *pg_encoding_to_char(int encoding);
+
+/* And this is the mod_sql_postgres-specific function mapping iconv
+ * locales to Postgres encodings.
+ */
+static const char *get_postgres_encoding(const char *encoding);
+#endif
+
 /* 
  * timer-handling code adds the need for a couple of forward declarations
  */
-MODRET cmd_close( cmd_rec *cmd );
+MODRET cmd_close(cmd_rec *cmd);
 module sql_postgres_module;
 
 /* 
@@ -94,8 +112,8 @@ typedef struct conn_entry_struct conn_entry_t;
 
 #define DEF_CONN_POOL_SIZE 10
 
-pool *conn_pool = NULL;
-array_header *conn_cache = NULL;
+static pool *conn_pool = NULL;
+static array_header *conn_cache = NULL;
 
 /*
  *  _sql_get_connection: walks the connection cache looking for the named
@@ -140,7 +158,7 @@ static void *_sql_add_connection(pool *p, char *name, db_conn_t *conn)
     return NULL;
   }
 
-  entry = (conn_entry_t *) pcalloc( p, sizeof( conn_entry_t ));
+  entry = (conn_entry_t *) pcalloc(p, sizeof(conn_entry_t));
   entry->name = name;
   entry->data = conn;
 
@@ -153,32 +171,31 @@ static void *_sql_add_connection(pool *p, char *name, db_conn_t *conn)
  *  properly filled in.  If not, it's grounds for the daemon to
  *  shutdown.
  */
-static void _sql_check_cmd(cmd_rec *cmd, char *msg)
-{
-  if ((!cmd) || (!cmd->tmp_pool)) {
-    log_pri(PR_LOG_ERR, _MOD_VERSION ": '%s' was passed an invalid cmd_rec. "
-	    "Shutting down.", msg);
-    sql_log(DEBUG_WARN, "'%s' was passed an invalid cmd_rec. Shutting down.",
+static void _sql_check_cmd(cmd_rec *cmd, char *msg) {
+  if (cmd == NULL ||
+      cmd->tmp_pool == NULL) {
+    pr_log_pri(PR_LOG_ERR, MOD_SQL_POSTGRES_VERSION
+      ": '%s' was passed an invalid cmd_rec (internal bug); shutting down",
       msg);
-    end_login(1);
+    sql_log(DEBUG_WARN, "'%s' was passed an invalid cmd_rec (internal bug); "
+      "shutting down", msg);
+    pr_session_end(0);
   }    
 
   return;
 }
 
 /*
- * _sql_timer_callback: when a timer goes off, this is the function
- *  that gets called.  This function makes assumptions about the 
- *  db_conn_t members.
+ * sql_timer_cb: when a timer goes off, this is the function that gets called.
+ * This function makes assumptions about the db_conn_t members.
  */
-static int _sql_timer_callback(CALLBACK_FRAME)
-{
+static int sql_timer_cb(CALLBACK_FRAME) {
   conn_entry_t *entry = NULL;
-  int cnt = 0;
+  int i = 0;
   cmd_rec *cmd = NULL;
  
-  for (cnt=0; cnt < conn_cache->nelts; cnt++) {
-    entry = ((conn_entry_t **) conn_cache->elts)[cnt];
+  for (i = 0; i < conn_cache->nelts; i++) {
+    entry = ((conn_entry_t **) conn_cache->elts)[i];
 
     if (entry->timer == p2) {
       sql_log(DEBUG_INFO, "timer expired for connection '%s'", entry->name);
@@ -192,40 +209,17 @@ static int _sql_timer_callback(CALLBACK_FRAME)
   return 0;
 }
 
-/* 
- * _sql_shutdown: walks the connection cache and closes every
- *  open connection, resetting their connection counts to 0.
- */
-static void _sql_shutdown(void)
-{
-  conn_entry_t *entry = NULL;
-  int cnt = 0;
-  cmd_rec *cmd;
-
-  for (cnt=0; cnt < conn_cache->nelts; cnt++) {
-    entry = ((conn_entry_t **) conn_cache->elts)[cnt];
-
-    if (entry->connections > 0) {
-      cmd = _sql_make_cmd( conn_pool, 2, entry->name, "1" );
-      cmd_close( cmd );
-      SQL_FREE_CMD(cmd);
-    }
-  }
-
-  return;
-}
-
 /* 
  * _build_error: constructs a modret_t filled with error information;
  *  mod_sql_postgres calls this function and returns the resulting mod_ret_t
  *  whenever a call to the database results in an error.
  */
-static modret_t *_build_error( cmd_rec *cmd, db_conn_t *conn )
-{
+static modret_t *_build_error(cmd_rec *cmd, db_conn_t *conn) {
   if (!conn)
-    return PR_ERR_SQL_BADCMD(cmd);
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "badly formed request");
 
-  return ERROR_MSG(cmd, "unknown", PQerrorMessage(conn->postgres));
+  return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION,
+    PQerrorMessage(conn->postgres));
 }
 
 /*
@@ -234,8 +228,7 @@ static modret_t *_build_error( cmd_rec *cmd, db_conn_t *conn )
  *  that data.  This is Postgres specific; other backends may choose 
  *  to do things differently.
  */
-static modret_t *_build_data( cmd_rec *cmd, db_conn_t *conn )
-{
+static modret_t *_build_data(cmd_rec *cmd, db_conn_t *conn) {
   PGresult *result = NULL;
   sql_data_t *sd = NULL;
   char **data = NULL;
@@ -244,7 +237,7 @@ static modret_t *_build_data( cmd_rec *cmd, db_conn_t *conn )
   int row =0;
 
   if (!conn) 
-    return PR_ERR_SQL_BADCMD(cmd);
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "badly formed request");
 
   result = conn->result;
 
@@ -252,8 +245,8 @@ static modret_t *_build_data( cmd_rec *cmd, db_conn_t *conn )
   sd->rnum = (unsigned long) PQntuples(result);
   sd->fnum = (unsigned long) PQnfields(result);
 
-  data = (char **) pcalloc( cmd->tmp_pool, sizeof(char *) * 
-			    ((sd->rnum * sd->fnum) + 1) );
+  data = (char **) pcalloc(cmd->tmp_pool, sizeof(char *) * 
+			    ((sd->rnum * sd->fnum) + 1));
   
   for (row = 0; row < sd->rnum; row++) {
     for (field = 0; field < sd->fnum; field++) {
@@ -267,6 +260,86 @@ static modret_t *_build_data( cmd_rec *cmd, db_conn_t *conn )
   return mod_create_data( cmd, (void *) sd );
 }
 
+#ifdef PR_USE_NLS
+static const char *get_postgres_encoding(const char *encoding) {
+
+  /* XXX Hack to deal with Postgres' incredibly broken behavior when
+   * handling the 'ASCII' encoding.  Specifically, Postgres chokes on
+   * 'ASCII', and instead insists on calling it 'SQL_ASCII' (which, of
+   * course, is not even close to being a valid encoding name according
+   * to libiconv.)
+   *
+   * Treat 'ANSI_X3.4-1968' (another common name/alias for ASCII) the
+   * same; rename it to 'SQL_ASCII' for Postgres' benefit.  And same for
+   * 'US-ASCII'.
+   */
+  if (strcasecmp(encoding, "ANSI_X3.4-1968") == 0 ||
+      strcasecmp(encoding, "ASCII") == 0 ||
+      strcasecmp(encoding, "US-ASCII") == 0) {
+    return "SQL_ASCII";
+  }
+
+  /* And other commonly used charsets for which Postgres has their own names. */
+
+  if (strcasecmp(encoding, "CP1251") == 0 ||
+      strcasecmp(encoding, "WINDOWS-1251") == 0) {
+    return "WIN1251";
+  }
+
+  if (strcasecmp(encoding, "KOI-8") == 0 ||
+      strcasecmp(encoding, "KOI8-R") == 0 ||
+      strcasecmp(encoding, "KOI8") == 0 ||
+      strcasecmp(encoding, "KOI8R") == 0) {
+    return "KOI";
+  }
+
+  if (strcasecmp(encoding, "CP866") == 0) {
+    return "WIN866";
+  }
+
+  if (strcasecmp(encoding, "ISO-8859-1") == 0) {
+    return "LATIN1";
+  }
+
+  if (strcasecmp(encoding, "ISO-8859-15") == 0) {
+    return "LATIN9";
+  }
+
+  if (strcasecmp(encoding, "EUC-CN") == 0 ||
+      strcasecmp(encoding, "EUCCN") == 0) {
+    return "EUC_CN";
+  }
+
+  if (strcasecmp(encoding, "EUC-JP") == 0 ||
+      strcasecmp(encoding, "EUCJP") == 0) {
+    return "EUC_JP";
+  }
+
+  if (strcasecmp(encoding, "EUC-KR") == 0 ||
+      strcasecmp(encoding, "EUCKR") == 0) {
+    return "EUC_KR";
+  }
+
+  if (strcasecmp(encoding, "EUC-TW") == 0 ||
+      strcasecmp(encoding, "EUCTW") == 0) {
+    return "EUC_TW";
+  }
+
+  if (strcasecmp(encoding, "SHIFT-JIS") == 0 ||
+      strcasecmp(encoding, "SHIFT_JIS") == 0) {
+    return "SJIS";
+  }
+
+  if (strcasecmp(encoding, "UTF8") == 0 ||
+      strcasecmp(encoding, "UTF-8") == 0 ||
+      strcasecmp(encoding, "UTF8-MAC") == 0) {
+    return "UTF8";
+  }
+
+  return encoding;
+}
+#endif
+
 /*
  * cmd_open: attempts to open a named connection to the database.
  *
@@ -284,10 +357,10 @@ static modret_t *_build_data( cmd_rec *cmd, db_conn_t *conn )
  *  compatibility; a connection should not be closed unless the count
  *  reaches 0, and ideally will not need to be re-opened for counts > 1.
  */
-MODRET cmd_open(cmd_rec *cmd)
-{
+MODRET cmd_open(cmd_rec *cmd) {
   conn_entry_t *entry = NULL;
   db_conn_t *conn = NULL;
+  const char *server_version = NULL;
 
   sql_log(DEBUG_FUNC, "%s", "entering \tpostgres cmd_open");
 
@@ -295,14 +368,15 @@ MODRET cmd_open(cmd_rec *cmd)
 
   if (cmd->argc < 1) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_open");
-    return PR_ERR_SQL_BADCMD(cmd);
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "badly formed request");
   }    
 
   /* get the named connection */
 
-  if (!(entry = _sql_get_connection( cmd->argv[0]))) {
+  if (!(entry = _sql_get_connection(cmd->argv[0]))) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_open");
-    return PR_ERR_SQL_UNDEF(cmd);
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION,
+      "unknown named connection");
   } 
 
   conn = (db_conn_t *) entry->data;
@@ -310,16 +384,77 @@ MODRET cmd_open(cmd_rec *cmd)
   /* if we're already open (connections > 0) increment connections 
    * reset our timer if we have one, and return HANDLED 
    */
-  if ((entry->connections > 0) && 
-      (PQstatus(conn->postgres) == CONNECTION_OK)) {
-    entry->connections++;
-    if (entry->timer) {
-      reset_timer( entry->timer, &sql_postgres_module );
+  if (entry->connections > 0) { 
+    if (PQstatus(conn->postgres) == CONNECTION_OK) {
+      entry->connections++;
+
+      if (entry->timer) {
+        pr_timer_reset(entry->timer, &sql_postgres_module);
+      }
+
+      sql_log(DEBUG_INFO, "connection '%s' count is now %d", entry->name,
+        entry->connections);
+      sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_open");
+      return PR_HANDLED(cmd);
+
+    } else {
+      char *reason;
+      size_t reason_len;
+
+      /* Unless we've been told not to reconnect, try to reconnect now.
+       * We only try once; if it fails, we return an error.
+       */
+      if (!(pr_sql_opts & SQL_OPT_NO_RECONNECT)) {
+        PQreset(conn->postgres);
+
+        if (PQstatus(conn->postgres) == CONNECTION_OK) {
+          entry->connections++;
+
+          if (entry->timer) {
+            pr_timer_reset(entry->timer, &sql_postgres_module);
+          }
+
+          sql_log(DEBUG_INFO, "connection '%s' count is now %d", entry->name,
+            entry->connections);
+          sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_open");
+          return PR_HANDLED(cmd);
+        }
+      }
+
+      reason = PQerrorMessage(conn->postgres);
+      reason_len = strlen(reason);
+
+      /* Postgres might give us an empty string as the reason; not helpful. */
+      if (reason_len == 0) {
+        reason = "(unknown)";
+        reason_len = strlen(reason);
+      }
+
+      /* The error message returned by Postgres is usually appended with
+       * a newline.  Let's prettify it by removing the newline.  Note
+       * that yes, we are overwriting the pointer given to us by Postgres,
+       * but it's OK.  The Postgres docs say that we're not supposed to
+       * free the memory associated with the returned string anyway.
+       */
+      reason = pstrdup(session.pool, reason);
+
+      if (reason[reason_len-1] == '\n') {
+        reason[reason_len-1] = '\0';
+        reason_len--;
+      }
+
+      sql_log(DEBUG_INFO, "lost connection to database: %s", reason);
+
+      entry->connections = 0;
+      if (entry->timer) {
+        pr_timer_remove(entry->timer, &sql_postgres_module);
+        entry->timer = 0;
+      }
+
+      sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_open");
+      return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION,
+        "lost connection to database");
     }
-    sql_log(DEBUG_INFO, "connection '%s' count is now %d", entry->name,
-      entry->connections);
-    sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_open");
-    return HANDLED(cmd);
   }
 
   /* make sure we have a new conn struct */
@@ -331,18 +466,56 @@ MODRET cmd_open(cmd_rec *cmd)
     return _build_error( cmd, conn );
   }
 
+#if defined(PG_VERSION_STR)
+  sql_log(DEBUG_FUNC, "Postgres client: %s", PG_VERSION_STR);
+#endif
+
+  server_version = PQparameterStatus(conn->postgres, "server_version");
+  if (server_version != NULL) {
+    sql_log(DEBUG_FUNC, "Postgres server version: %s", server_version);
+  }
+
+#ifdef PR_USE_NLS
+  if (pr_encode_get_encoding() != NULL) {
+    const char *encoding;
+
+    encoding = get_postgres_encoding(pr_encode_get_encoding());
+
+    /* Configure the connection for the current local character set. */
+    if (PQsetClientEncoding(conn->postgres, encoding) < 0) {
+      /* if it didn't work, return an error */
+      sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_open");
+      return _build_error(cmd, conn);
+    }
+
+    sql_log(DEBUG_FUNC, "Postgres connection character set now '%s' "
+      "(from '%s')", pg_encoding_to_char(PQclientEncoding(conn->postgres)),
+      pr_encode_get_encoding());
+  }
+#endif /* !PR_USE_NLS */
+
   /* bump connections */
   entry->connections++;
 
-  /* set up our timer if necessary */
-  if (entry->ttl > 0) {
-    entry->timer = add_timer(entry->ttl, -1, 
-			     &sql_postgres_module, 
-			     _sql_timer_callback);
+  if (pr_sql_conn_policy == SQL_CONN_POLICY_PERSESSION) {
+    /* If the connection policy is PERSESSION... */
+    if (entry->connections == 1) {
+      /* ...and we are actually opening the first connection to the database;
+       * we want to make sure this connection stays open, after this first use
+       * (as per Bug#3290).  To do this, we re-bump the connection count.
+       */
+      entry->connections++;
+    } 
+ 
+  } else if (entry->ttl > 0) { 
+    /* Set up our timer if necessary */
+
+    entry->timer = pr_timer_add(entry->ttl, -1, &sql_postgres_module,
+      sql_timer_cb, "postgres connection ttl");
     sql_log(DEBUG_INFO, "connection '%s' - %d second timer started",
       entry->name, entry->ttl);
 
-    /* timed connections get re-bumped so they don't go away when cmd_close
+    /* Timed connections get re-bumped so they don't go away when cmd_close
      * is called.
      */
     entry->connections++;
@@ -355,7 +528,7 @@ MODRET cmd_open(cmd_rec *cmd)
     entry->connections);
 
   sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_open");
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 /*
@@ -382,8 +555,7 @@ MODRET cmd_open(cmd_rec *cmd)
  *  If argv[1] exists and is not NULL, the connection should be immediately
  *  closed and the connection count should be reset to 0.
  */
-MODRET cmd_close(cmd_rec *cmd)
-{
+MODRET cmd_close(cmd_rec *cmd) {
   conn_entry_t *entry = NULL;
   db_conn_t *conn = NULL;
 
@@ -393,13 +565,14 @@ MODRET cmd_close(cmd_rec *cmd)
 
   if ((cmd->argc < 1) || (cmd->argc > 2)) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_close");
-    return PR_ERR_SQL_BADCMD(cmd);
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "badly formed request");
   }
 
   /* get the named connection */
-  if (!(entry = _sql_get_connection( cmd->argv[0] ))) {
+  if (!(entry = _sql_get_connection(cmd->argv[0]))) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_close");
-    return PR_ERR_SQL_UNDEF(cmd);
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION,
+      "unknown named connection");
   }
 
   conn = (db_conn_t *) entry->data;
@@ -410,7 +583,7 @@ MODRET cmd_close(cmd_rec *cmd)
       entry->connections);
 
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_close");
-    return HANDLED(cmd);
+    return PR_HANDLED(cmd);
   }
 
   /* decrement connections. If our count is 0 or we received a second arg
@@ -423,7 +596,7 @@ MODRET cmd_close(cmd_rec *cmd)
     entry->connections = 0;
 
     if (entry->timer) {
-      remove_timer( entry->timer, &sql_postgres_module );
+      pr_timer_remove(entry->timer, &sql_postgres_module);
       entry->timer = 0;
       sql_log(DEBUG_INFO, "connection '%s' - timer stopped", entry->name);
     }
@@ -435,7 +608,7 @@ MODRET cmd_close(cmd_rec *cmd)
     entry->connections);
   sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_close");
   
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 /*
@@ -461,8 +634,7 @@ MODRET cmd_close(cmd_rec *cmd)
  *  a number or ttl is negative, the connection will be assumed to have no
  *  associated timer.
  */
-MODRET cmd_defineconnection(cmd_rec *cmd)
-{
+MODRET cmd_defineconnection(cmd_rec *cmd) {
   char *info = NULL;
   char *name = NULL;
 
@@ -483,10 +655,24 @@ MODRET cmd_defineconnection(cmd_rec *cmd)
 
   if ((cmd->argc < 4) || (cmd->argc > 5) || (!cmd->argv[0])) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_defineconnection");
-    return PR_ERR_SQL_BADCMD(cmd);
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "badly formed request");
   }
 
-  conn = (db_conn_t *) palloc(conn_pool, sizeof(db_conn_t));
+  if (!conn_pool) {
+    pr_log_pri(PR_LOG_WARNING, "warning: the mod_sql_postgres module has not "
+      "been properly initialized.  Please make sure your --with-modules "
+      "configure option lists mod_sql *before* mod_sql_postgres, and "
+      "recompile.");
+      
+    sql_log(DEBUG_FUNC, "%s", "The mod_sql_postgres module has not been "
+      "properly initialized.  Please make sure your --with-modules configure "
+      "option lists mod_sql *before* mod_sql_postgres, and recompile.");
+    sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_defineconnection");
+
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "uninitialized module");
+  }
+
+  conn = (db_conn_t *) pcalloc(conn_pool, sizeof(db_conn_t));
 
   name = pstrdup(conn_pool, cmd->argv[0]);
   conn->user = pstrdup(conn_pool, cmd->argv[1]);
@@ -535,13 +721,19 @@ MODRET cmd_defineconnection(cmd_rec *cmd)
   /* insert the new conn_info into the connection hash */
   if (!(entry = _sql_add_connection(conn_pool, name, (void *) conn))) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_defineconnection");
-    return PR_ERR_SQL_REDEF(cmd);
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION,
+      "named connection already exists");
   }
 
-  entry->ttl = (cmd->argc == 5) ? 
-    (int) strtol(cmd->argv[4], (char **)NULL, 10) : 0;
-  if (entry->ttl < 0) 
-    entry->ttl = 0;
+  if (cmd->argc == 5) { 
+    entry->ttl = (int) strtol(cmd->argv[4], (char **) NULL, 10);
+    if (entry->ttl >= 1) {
+      pr_sql_conn_policy = SQL_CONN_POLICY_TIMER;
+ 
+    } else {
+      entry->ttl = 0;
+    }
+  }
 
   entry->timer = 0;
   entry->connections = 0;
@@ -554,7 +746,36 @@ MODRET cmd_defineconnection(cmd_rec *cmd)
   sql_log(DEBUG_INFO, "  ttl: '%d'", entry->ttl);
 
   sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_defineconnection");
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
+}
+
+/*
+ * cmd_exit: closes all open connections.
+ *
+ * Inputs:
+ *  None
+ *
+ * Returns:
+ *  A simple non-error modret_t.
+ */
+static modret_t *cmd_exit(cmd_rec *cmd) {
+  register unsigned int i = 0;
+
+  sql_log(DEBUG_FUNC, "%s", "entering \tpostgres cmd_exit");
+
+  for (i = 0; i < conn_cache->nelts; i++) {
+    conn_entry_t *entry = ((conn_entry_t **) conn_cache->elts)[i];
+
+    if (entry->connections > 0) {
+      cmd_rec *close_cmd = _sql_make_cmd(conn_pool, 2, entry->name, "1");
+      cmd_close(close_cmd);
+      destroy_pool(close_cmd->pool);
+    }
+  }
+
+  sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_exit");
+
+  return PR_HANDLED(cmd);
 }
 
 /*
@@ -603,8 +824,7 @@ MODRET cmd_defineconnection(cmd_rec *cmd)
  *  certain selects could return huge amounts of data.  do whatever is
  *  possible to minimize the amount of data copying here.
  */
-MODRET cmd_select(cmd_rec *cmd)
-{
+MODRET cmd_select(cmd_rec *cmd) {
   conn_entry_t *entry = NULL;
   db_conn_t *conn = NULL;
   modret_t *cmr = NULL;
@@ -619,14 +839,15 @@ MODRET cmd_select(cmd_rec *cmd)
 
   if (cmd->argc < 2) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_select");
-    return PR_ERR_SQL_BADCMD(cmd);
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "badly formed request");
   }
 
   /* get the named connection */
-  entry = _sql_get_connection( cmd->argv[0] );
+  entry = _sql_get_connection(cmd->argv[0]);
   if (!entry) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_select");
-    return PR_ERR_SQL_UNDEF(cmd);
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION,
+      "unknown named connection");
   }
   
   conn = (db_conn_t *) entry->data;
@@ -742,8 +963,7 @@ MODRET cmd_select(cmd_rec *cmd)
  * Notes:
  *  none
  */
-MODRET cmd_insert(cmd_rec *cmd)
-{
+MODRET cmd_insert(cmd_rec *cmd) {
   conn_entry_t *entry = NULL;
   db_conn_t *conn = NULL;
   modret_t *cmr = NULL;
@@ -757,14 +977,15 @@ MODRET cmd_insert(cmd_rec *cmd)
 
   if ((cmd->argc != 2) && (cmd->argc != 4)) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_insert");
-    return PR_ERR_SQL_BADCMD(cmd);
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "badly formed request");
   }
 
   /* get the named connection */
-  entry = _sql_get_connection( cmd->argv[0] );
+  entry = _sql_get_connection(cmd->argv[0]);
   if (!entry) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_insert");
-    return PR_ERR_SQL_UNDEF(cmd);
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION,
+      "unknown named connection");
   }
 
   conn = (db_conn_t *) entry->data;
@@ -812,7 +1033,7 @@ MODRET cmd_insert(cmd_rec *cmd)
   SQL_FREE_CMD(close_cmd);
 
   sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_insert");
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 /*
@@ -845,8 +1066,7 @@ MODRET cmd_insert(cmd_rec *cmd)
  *  argv[3] is optional -- it may be NULL, or it may not exist at all.  
  *  make sure this is handled correctly. 
  */
-MODRET cmd_update(cmd_rec *cmd)
-{
+MODRET cmd_update(cmd_rec *cmd) {
   conn_entry_t *entry = NULL;
   db_conn_t *conn = NULL;
   modret_t *cmr = NULL;
@@ -860,14 +1080,15 @@ MODRET cmd_update(cmd_rec *cmd)
 
   if ((cmd->argc < 2) || (cmd->argc > 4)) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_update");
-    return PR_ERR_SQL_BADCMD(cmd);
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "badly formed request");
   }
 
   /* get the named connection */
-  entry = _sql_get_connection( cmd->argv[0] );
+  entry = _sql_get_connection(cmd->argv[0]);
   if (!entry) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_update");
-    return PR_ERR_SQL_UNDEF(cmd);
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION,
+      "unknown named connection");
   }
 
   conn = (db_conn_t *) entry->data;
@@ -916,7 +1137,7 @@ MODRET cmd_update(cmd_rec *cmd)
   SQL_FREE_CMD(close_cmd);
 
   sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_update");
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 /*
@@ -937,22 +1158,22 @@ MODRET cmd_update(cmd_rec *cmd)
  *  not support stored procedures should return an error with a descriptive
  *  error message (something like 'backend does not support procedures').
  */
-MODRET cmd_procedure(cmd_rec *cmd)
-{
+MODRET cmd_procedure(cmd_rec *cmd) {
   sql_log(DEBUG_FUNC, "%s", "entering \tpostgres cmd_procedure");
 
   _sql_check_cmd(cmd, "cmd_procedure");
 
   if (cmd->argc != 3) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_procedure");
-    return PR_ERR_SQL_BADCMD(cmd);
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "badly formed request");
   }
 
   /* PostgreSQL supports procedures, but the backend doesn't. */
 
   sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_procedure");
 
-  return ERROR_MSG(cmd, _MOD_VERSION, "backend does not support procedures");
+  return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION,
+    "backend does not support procedures");
 }
 
 /*
@@ -974,8 +1195,7 @@ MODRET cmd_procedure(cmd_rec *cmd)
  * Notes:
  *  None.
  */
-MODRET cmd_query(cmd_rec *cmd)
-{
+MODRET cmd_query(cmd_rec *cmd) {
   conn_entry_t *entry = NULL;
   db_conn_t *conn = NULL;
   modret_t *cmr = NULL;
@@ -989,14 +1209,15 @@ MODRET cmd_query(cmd_rec *cmd)
 
   if (cmd->argc != 2) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_query");
-    return PR_ERR_SQL_BADCMD(cmd);
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "badly formed request");
   }
 
   /* get the named connection */
-  entry = _sql_get_connection( cmd->argv[0] );
+  entry = _sql_get_connection(cmd->argv[0]);
   if (!entry) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_query");
-    return PR_ERR_SQL_UNDEF(cmd);
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION,
+      "unknown named connection");
   }
 
   conn = (db_conn_t *) entry->data;
@@ -1043,7 +1264,7 @@ MODRET cmd_query(cmd_rec *cmd)
       sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_query");
     }
   } else {
-    dmr = HANDLED(cmd);
+    dmr = PR_HANDLED(cmd);
   }
 
   /* close the connection, return the data. */
@@ -1084,8 +1305,14 @@ MODRET cmd_query(cmd_rec *cmd)
 MODRET cmd_escapestring(cmd_rec * cmd) {
   conn_entry_t *entry = NULL;
   db_conn_t *conn = NULL;
+  modret_t *cmr = NULL;
   char *unescaped = NULL;
   char *escaped = NULL;
+  cmd_rec *close_cmd;
+  size_t unescaped_len = 0;
+#ifdef HAVE_POSTGRES_PQESCAPESTRINGCONN
+  int pgerr = 0;
+#endif
 
   sql_log(DEBUG_FUNC, "%s", "entering \tpostgres cmd_escapestring");
 
@@ -1093,26 +1320,44 @@ MODRET cmd_escapestring(cmd_rec * cmd) {
 
   if (cmd->argc != 2) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_escapestring");
-    return PR_ERR_SQL_BADCMD(cmd);
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "badly formed request");
   }
 
   /* get the named connection */
-  entry = _sql_get_connection( cmd->argv[0] );
+  entry = _sql_get_connection(cmd->argv[0]);
   if (!entry) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_escapestring");
-    return PR_ERR_SQL_UNDEF(cmd);
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION,
+      "unknown named connection");
   }
 
   conn = (db_conn_t *) entry->data;
 
-  /* Note: the PQescapeString() function appeared in the C API as of
-   * Postgres-7.2.
-   */
-  unescaped = cmd->argv[1];
-  escaped = (char *) pcalloc(cmd->tmp_pool, sizeof(char) *
-    (strlen(unescaped) * 2) + 1);
+  /* Make sure the connection is open. */
+  cmr = cmd_open(cmd);
+  if (MODRET_ERROR(cmr)) {
+    sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_escapestring");
+    return cmr;
+  }
 
-  PQescapeString(escaped, unescaped, strlen(unescaped));
+  unescaped = cmd->argv[1];
+  unescaped_len = strlen(unescaped);
+  escaped = (char *) pcalloc(cmd->tmp_pool, sizeof(char) *
+    (unescaped_len * 2) + 1);
+
+#ifdef HAVE_POSTGRES_PQESCAPESTRINGCONN
+  PQescapeStringConn(conn->postgres, escaped, unescaped, unescaped_len, &pgerr);
+  if (pgerr != 0) {
+    sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_escapestring");
+    return _build_error(cmd, conn);
+  }
+#else
+  PQescapeString(escaped, unescaped, unescaped_len);
+#endif
+
+  close_cmd = _sql_make_cmd(cmd->tmp_pool, 1, entry->name);
+  cmd_close(close_cmd);
+  SQL_FREE_CMD(close_cmd);
 
   sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_escapestring");
   return mod_create_data(cmd, (void *) escaped);
@@ -1129,19 +1374,18 @@ MODRET cmd_escapestring(cmd_rec * cmd) {
  *  cmd->argv[2]: hashed string
  *
  * Returns:
- *  HANDLED(cmd)                        -- passwords match
- *  ERROR_INT(cmd, PR_AUTH_NOPWD)       -- missing password
- *  ERROR_INT(cmd, PR_AUTH_BADPWD)      -- passwords don't match
- *  ERROR_INT(cmd, PR_AUTH_DISABLEPWD)  -- password is disabled
- *  ERROR_INT(cmd, PR_AUTH_AGEPWD)      -- password is aged
- *  ERROR(cmd)                          -- unknown error
+ *  PR_HANDLED(cmd)                        -- passwords match
+ *  PR_ERROR_INT(cmd, PR_AUTH_NOPWD)       -- missing password
+ *  PR_ERROR_INT(cmd, PR_AUTH_BADPWD)      -- passwords don't match
+ *  PR_ERROR_INT(cmd, PR_AUTH_DISABLEPWD)  -- password is disabled
+ *  PR_ERROR_INT(cmd, PR_AUTH_AGEPWD)      -- password is aged
+ *  PR_ERROR(cmd)                          -- unknown error
  *
  * Notes:
  *  If this backend does not provide this functionality, this cmd *must*
  *  return ERROR.
  */
-MODRET cmd_checkauth(cmd_rec * cmd)
-{
+MODRET cmd_checkauth(cmd_rec * cmd) {
   conn_entry_t *entry = NULL;
   db_conn_t *conn = NULL;
 
@@ -1151,23 +1395,27 @@ MODRET cmd_checkauth(cmd_rec * cmd)
 
   if (cmd->argc != 3) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_checkauth");
-    return PR_ERR_SQL_BADCMD(cmd);
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "badly formed request");
   }
 
   /* get the named connection -- not used in this case, but for consistency */
-  entry = _sql_get_connection( cmd->argv[0] );
+  entry = _sql_get_connection(cmd->argv[0]);
   if (!entry) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_checkauth");
-    return PR_ERR_SQL_UNDEF(cmd);
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION,
+      "unknown named connection");
   }
 
   conn = (db_conn_t *) entry->data;
 
+  sql_log(DEBUG_WARN, MOD_SQL_POSTGRES_VERSION
+    ": Postgres does not support the 'Backend' SQLAuthType");
+
   sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_checkauth");
 
   /* PostgreSQL doesn't provide this functionality */
-
-  return ERROR(cmd);
+  return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION,
+    "Postgres does not support the 'Backend' SQLAuthType");
 }
 
 /*
@@ -1195,53 +1443,121 @@ MODRET cmd_identify(cmd_rec * cmd) {
 
   _sql_check_cmd(cmd, "cmd_identify");
 
-  sd = (sql_data_t *) pcalloc( cmd->tmp_pool, sizeof(sql_data_t));
-  sd->data = (char **) pcalloc( cmd->tmp_pool, sizeof(char *) * 2);
+  sd = (sql_data_t *) pcalloc(cmd->tmp_pool, sizeof(sql_data_t));
+  sd->data = (char **) pcalloc(cmd->tmp_pool, sizeof(char *) * 2);
 
   sd->rnum = 1;
   sd->fnum = 2;
 
-  sd->data[0] = _MOD_VERSION;
+  sd->data[0] = MOD_SQL_POSTGRES_VERSION;
   sd->data[1] = MOD_SQL_API_V1;
 
   return mod_create_data(cmd, (void *) sd);
 }  
 
-/* 
- * sql_cmdtable: mod_sql requires each backend module to define a cmdtable
+MODRET cmd_prepare(cmd_rec *cmd) {
+  if (cmd->argc != 1) {
+    return PR_ERROR(cmd);
+  }
+
+  conn_pool = (pool *) cmd->argv[0];
+
+  if (conn_cache == NULL) {
+    conn_cache = make_array((pool *) cmd->argv[0], DEF_CONN_POOL_SIZE,
+      sizeof(conn_entry_t *));
+  }
+
+  return mod_create_data(cmd, NULL);
+}
+
+MODRET cmd_cleanup(cmd_rec *cmd) {
+  destroy_pool(conn_pool);
+  conn_pool = NULL;
+  conn_cache = NULL;
+
+  return mod_create_data(cmd, NULL);
+}
+
+/* SQL cmdtable: mod_sql requires each backend module to define a cmdtable
  *  with this exact name. ALL these functions must be defined; mod_sql checks
  *  that they all exist on startup and ProFTPD will refuse to start if they
  *  aren't defined.
  */
-cmdtable sql_cmdtable[] = {
-  { CMD, "sql_open",             G_NONE, cmd_open,             FALSE, FALSE },
+static cmdtable sql_postgres_cmdtable[] = {
+  { CMD, "sql_checkauth",        G_NONE, cmd_checkauth,        FALSE, FALSE },
+  { CMD, "sql_cleanup",          G_NONE, cmd_cleanup,          FALSE, FALSE },
   { CMD, "sql_close",            G_NONE, cmd_close,            FALSE, FALSE },
   { CMD, "sql_defineconnection", G_NONE, cmd_defineconnection, FALSE, FALSE },
-  { CMD, "sql_select",           G_NONE, cmd_select,           FALSE, FALSE },
+  { CMD, "sql_escapestring",     G_NONE, cmd_escapestring,     FALSE, FALSE },
+  { CMD, "sql_exit",             G_NONE, cmd_exit,             FALSE, FALSE },
+  { CMD, "sql_identify",         G_NONE, cmd_identify,         FALSE, FALSE },
   { CMD, "sql_insert",           G_NONE, cmd_insert,           FALSE, FALSE },
-  { CMD, "sql_update",           G_NONE, cmd_update,           FALSE, FALSE },
+  { CMD, "sql_open",             G_NONE, cmd_open,             FALSE, FALSE },
+  { CMD, "sql_prepare",          G_NONE, cmd_prepare,          FALSE, FALSE },
   { CMD, "sql_procedure",        G_NONE, cmd_procedure,        FALSE, FALSE },
   { CMD, "sql_query",            G_NONE, cmd_query,            FALSE, FALSE },
-  { CMD, "sql_escapestring",     G_NONE, cmd_escapestring,     FALSE, FALSE },
-  { CMD, "sql_checkauth",        G_NONE, cmd_checkauth,        FALSE, FALSE },
-  { CMD, "sql_identify",         G_NONE, cmd_identify,         FALSE, FALSE },
+  { CMD, "sql_select",           G_NONE, cmd_select,           FALSE, FALSE },
+  { CMD, "sql_update",           G_NONE, cmd_update,           FALSE, FALSE },
 
   { 0, NULL }
 };
 
-/*
- * sql_postgres_init: Used to initialize the connection cache and register
- *  the exit handler.
+/* Event handlers
  */
+
+static void sql_postgres_mod_load_ev(const void *event_data,
+    void *user_data) {
+
+  if (strcmp("mod_sql_postgres.c", (const char *) event_data) == 0) {
+    /* Register ourselves with mod_sql. */
+    if (sql_register_backend("postgres", sql_postgres_cmdtable) < 0) {
+      pr_log_pri(PR_LOG_NOTICE, MOD_SQL_POSTGRES_VERSION
+        ": notice: error registering backend: %s", strerror(errno));
+      pr_session_end(0);
+    }
+  }
+}
+
+static void sql_postgres_mod_unload_ev(const void *event_data,
+    void *user_data) {
+
+  if (strcmp("mod_sql_postgres.c", (const char *) event_data) == 0) {
+    /* Unegister ourselves with mod_sql. */
+    if (sql_unregister_backend("postgres") < 0) {
+      pr_log_pri(PR_LOG_NOTICE, MOD_SQL_POSTGRES_VERSION
+        ": notice: error unregistering backend: %s", strerror(errno));
+      pr_session_end(0);
+    }
+
+    /* Unregister ourselves from all events. */
+    pr_event_unregister(&sql_postgres_module, NULL, NULL);
+  }
+}
+
+/* Initialization routines
+ */
+
 static int sql_postgres_init(void) {
-  if (!conn_pool)
+
+  /* Register listeners for the load and unload events. */
+  pr_event_register(&sql_postgres_module, "core.module-load",
+    sql_postgres_mod_load_ev, NULL);
+  pr_event_register(&sql_postgres_module, "core.module-unload",
+    sql_postgres_mod_unload_ev, NULL);
+
+  return 0;
+}
+
+static int sql_postgres_sess_init(void) {
+  if (conn_pool == NULL) {
     conn_pool = make_sub_pool(session.pool);
+    pr_pool_tag(conn_pool, "Postgres connection pool");
+  }
 
-  if (!conn_cache)
+  if (conn_cache == NULL) {
     conn_cache = make_array(make_sub_pool(session.pool), DEF_CONN_POOL_SIZE,
-      sizeof(conn_entry_t));
-
-  pr_exit_register_handler( _sql_shutdown );
+      sizeof(conn_entry_t *));
+  }
 
   return 0;
 }
@@ -1253,12 +1569,30 @@ static int sql_postgres_init(void) {
  *  to extend the init functions to initialize other internal variables.
  */
 module sql_postgres_module = {
-  NULL, NULL,                   /* Always NULL */
-  0x20,                         /* API Version 2.0 */
+  /* Always NULL */
+  NULL, NULL,
+
+  /* Module API version */
+  0x20,
+
+  /* Module name */
   "sql_postgres",
-  NULL,                         /* SQL configuration handler table */
-  NULL,                         /* SQL command handler table */
-  NULL,                         /* SQL authentication handler table */
-  NULL,                         /* Pre-fork "daemon mode" init */
-  sql_postgres_init             /* Post-fork "child mode" init */
+
+  /* Module configuration directive handlers */
+  NULL,
+
+  /* Module command handlers */
+  NULL,
+
+  /* Module authentication handlers */
+  NULL,
+
+  /* Module initialization */
+  sql_postgres_init,
+
+  /* Session initialization */
+  sql_postgres_sess_init,
+
+  /* Module version */
+  MOD_SQL_POSTGRES_VERSION
 };
