@@ -2,7 +2,7 @@
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
  * Copyright (c) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
- * Copyright (c) 2001, 2002, 2003, 2004 The ProFTPD Project team
+ * Copyright (c) 2001-2013 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -16,7 +16,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
  *
  * As a special exemption, Public Flood Software/MacGyver aka Habeeb J. Dihu
  * and other respective copyright holders give permission to link this program
@@ -25,7 +25,7 @@
  */
 
 /* Generic configuration and standard header file includes.
- * $Id: conf.h,v 1.50 2004/04/29 19:28:24 castaglia Exp $
+ * $Id: conf.h,v 1.92 2013/09/19 05:54:32 castaglia Exp $
  */
 
 #ifndef PR_CONF_H
@@ -38,6 +38,15 @@
 #include "version.h"
 #include "config.h"
 
+/* Manually override the socklen_t type on HP-UX 11 boxes.  For more
+ * details on why, see:
+ *  http://nagoya.apache.org/bugzilla/show_bug.cgi?id=16317
+ */
+#if defined(HPUX11) && !defined(_XOPEN_SOURCE_EXTENDED)
+# undef socklen_t
+# define socklen_t      int
+#endif
+
 #include "default_paths.h"
 
 #include <stdlib.h>
@@ -54,6 +63,10 @@
 # include <sys/types.h>
 #endif
 
+#ifdef HAVE_INTTYPES_H
+# include <inttypes.h>
+#endif
+
 #ifdef HAVE_UNISTD_H
 # include <unistd.h>
 #endif
@@ -62,6 +75,14 @@
 # include <limits.h>
 #endif
 
+#ifdef HAVE_SIGNAL_H
+# include <signal.h>
+#endif
+
+#ifdef HAVE_SYS_ACL_H
+# include <sys/acl.h>
+#endif
+
 #ifdef HAVE_SYS_SELECT_H
 # include <sys/select.h>
 #endif
@@ -119,6 +140,10 @@ char *strchr(),*strrchr();
 # include <sys/param.h>
 #endif
 
+#ifdef HAVE_SYS_RESOURCE_H
+# include <sys/resource.h>
+#endif
+
 #ifdef HAVE_SYS_SOCKET_H
 # include <sys/socket.h>
 #endif
@@ -216,13 +241,31 @@ char *strchr(),*strrchr();
  * then try termio.h
  */
 
-#ifdef HAVE_SYS_TERMIOS_H
-# include <sys/termios.h>
+#ifdef HAVE_TERMIOS_H
+# include <termios.h>
 #else
-# ifdef HAVE_SYS_TERMIO_H
-#  include <sys/termio.h>
-# endif /* HAVE_SYS_TERMIO_H */
-#endif /* HAVE_SYS_TERMIOS_H */
+# ifdef HAVE_SYS_TERMIOS_H
+#  include <sys/termios.h>
+# else
+#  ifdef HAVE_SYS_TERMIO_H
+#   include <sys/termio.h>
+#  endif /* HAVE_SYS_TERMIO_H */
+# endif /* HAVE_SYS_TERMIOS_H */
+#endif /* HAVE_TERMIOS_H */
+
+#ifdef PR_USE_NLS
+# ifdef HAVE_LIBINTL_H
+#  include <libintl.h>
+# endif
+# define _(str) dgettext("proftpd", str)
+# ifdef HAVE_LOCALE_H
+#  include <locale.h>
+# endif
+#else
+# define _(str) (str)
+# define textdomain(dir)
+# define bindtextdomain(pkg, dir)
+#endif /* PR_USE_NLS */
 
 /* The tunable options header needs to be included after all the system headers,
  * so that limits are picked up properly.
@@ -241,9 +284,17 @@ char *strchr(),*strrchr();
 /* AIX, when compiled using -D_NO_PROTO, lacks some prototypes without
  * which ProFTPD may do some funny (and not good) things.  Provide the
  * prototypes as necessary here.
+ *
+ * As examples of what these "not good" things might be:
+ *
+ *  1.  The ScoreboardFile will grow endlessly; session slots will not
+ *      be cleared and reused properly.
+ *
+ *  2.  The mod_delay module will complain of being unable to load the
+ *      table into memory due to "Invalid argument".
  */
 
-#if defined(AIX4) && defined(_NO_PROTO)
+#if defined(_NO_PROTO) && (defined(AIX4) || defined(AIX5))
 off_t lseek(int, off_t, int);
 #endif
 
@@ -310,9 +361,11 @@ extern char *alloca();
 #endif
 
 #if defined(HAVE_LLU) && SIZEOF_OFF_T == 8
-# define PR_LU	"llu"
+# define PR_LU		"llu"
+# define pr_off_t	unsigned long long
 #else
-# define PR_LU	"lu"
+# define PR_LU		"lu"
+# define pr_off_t	unsigned long
 #endif
 
 /********************************************************************
@@ -343,11 +396,11 @@ typedef struct {
   char na_dnsstr[256];
   int na_have_dnsstr;
 
-#ifdef USE_IPV6
+#ifdef PR_USE_IPV6
   char na_ipstr[INET6_ADDRSTRLEN];
 #else
   char na_ipstr[INET_ADDRSTRLEN];
-#endif /* USE_IPV6 */
+#endif /* PR_USE_IPV6 */
   int na_have_ipstr;
 
   /* Note: at some point, this union might/should be replaced with
@@ -358,21 +411,30 @@ typedef struct {
 
   union {
     struct sockaddr_in v4;
-#ifdef USE_IPV6
+#ifdef PR_USE_IPV6
     struct sockaddr_in6 v6;
-#endif /* USE_IPV6 */
+#endif /* PR_USE_IPV6 */
   } na_addr;
 
 } pr_netaddr_t;
 
 #include "pool.h"
-#include "regexp.h"
+#include "str.h"
+#include "ascii.h"
+#include "table.h"
 #include "proftpd.h"
 #include "support.h"
+#include "str.h"
 #include "sets.h"
 #include "dirtree.h"
-#include "netio.h"
+#include "expr.h"
+#include "rlimit.h"
+#include "filter.h"
 #include "modules.h"
+#include "netio.h"
+#include "regexp.h"
+#include "stash.h"
+#include "auth.h"
 #include "response.h"
 #include "timers.h"
 #include "inet.h"
@@ -380,21 +442,33 @@ typedef struct {
 #include "netaddr.h"
 #include "netacl.h"
 #include "class.h"
+#include "cmd.h"
 #include "bindings.h"
-#include "ident.h"
 #include "help.h"
 #include "feat.h"
 #include "ftp.h"
 #include "log.h"
+#include "parser.h"
 #include "xferlog.h"
 #include "scoreboard.h"
 #include "data.h"
+#include "display.h"
 #include "libsupp.h"
 #include "fsio.h"
 #include "mkhome.h"
 #include "ctrls.h"
+#include "session.h"
 #include "event.h"
+#include "var.h"
+#include "throttle.h"
+#include "trace.h"
+#include "encode.h"
+#include "compat.h"
+#include "proctitle.h"
+#include "pidfile.h"
+#include "env.h"
 #include "pr-syslog.h"
+#include "memcache.h"
 
 # ifdef HAVE_SETPASSENT
 #  define setpwent()	setpassent(1)
