@@ -1,7 +1,7 @@
 /*
  * ProFTPD: mod_facts -- a module for handling "facts" [RFC3659]
  *
- * Copyright (c) 2007-2012 The ProFTPD Project
+ * Copyright (c) 2007-2013 The ProFTPD Project
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -22,7 +22,7 @@
  * resulting executable, without including the source code for OpenSSL in the
  * source distribution.
  *
- * $Id: mod_facts.c,v 1.45.2.4 2012/07/13 18:22:46 castaglia Exp $
+ * $Id: mod_facts.c,v 1.60 2013/04/16 16:04:43 castaglia Exp $
  */
 
 #include "conf.h"
@@ -61,6 +61,7 @@ struct mlinfo {
 
 /* Necessary prototypes */
 static void facts_mlinfobuf_flush(void);
+static int facts_sess_init(void);
 
 /* Support functions
  */
@@ -293,7 +294,7 @@ static size_t facts_mlinfo_fmt(struct mlinfo *info, char *buf, size_t bufsz) {
  * channel, wherease MLSD's output is sent via a data transfer, much like
  * LIST or NLST.
  */
-static char *mlinfo_buf = NULL;
+static char *mlinfo_buf = NULL, *mlinfo_bufptr = NULL;
 static size_t mlinfo_bufsz = 0;
 static size_t mlinfo_buflen = 0;
 
@@ -306,6 +307,7 @@ static void facts_mlinfobuf_init(void) {
   }
 
   memset(mlinfo_buf, '\0', mlinfo_bufsz);
+  mlinfo_bufptr = mlinfo_buf;
   mlinfo_buflen = 0;
 }
 
@@ -322,7 +324,8 @@ static void facts_mlinfobuf_add(struct mlinfo *info) {
     (void) facts_mlinfobuf_flush();
   }
 
-  sstrcat(mlinfo_buf, buf, mlinfo_bufsz);
+  sstrcat(mlinfo_bufptr, buf, mlinfo_bufsz - mlinfo_buflen);
+  mlinfo_bufptr += buflen;
   mlinfo_buflen += buflen;
 }
 
@@ -330,12 +333,20 @@ static void facts_mlinfobuf_flush(void) {
   if (mlinfo_buflen > 0) {
     int res;
 
+    /* Make sure the ASCII flags are cleared from the session flags,
+     * so that the pr_data_xfer() function does not try to perform
+     * ASCII translation on this data.
+     */
+    session.sf_flags &= ~SF_ASCII_OVERRIDE;
+
     res = pr_data_xfer(mlinfo_buf, mlinfo_buflen);
     if (res < 0 &&
         errno != 0) {
       pr_log_debug(DEBUG3, MOD_FACTS_VERSION
         ": error transferring data: [%d] %s", errno, strerror(errno));
     }
+
+    session.sf_flags |= SF_ASCII_OVERRIDE;
   }
 
   facts_mlinfobuf_init();
@@ -348,8 +359,12 @@ static int facts_mlinfo_get(struct mlinfo *info, const char *path,
 
   res = pr_fsio_lstat(path, &(info->st));
   if (res < 0) {
+    int xerrno = errno;
+
     pr_log_debug(DEBUG4, MOD_FACTS_VERSION ": error lstat'ing '%s': %s",
-      path, strerror(errno));
+      path, strerror(xerrno));
+
+    errno = xerrno;
     return -1;
   }
 
@@ -376,8 +391,12 @@ static int facts_mlinfo_get(struct mlinfo *info, const char *path,
       pr_fs_clear_cache();
       res = pr_fsio_stat(path, &target_st);
       if (res < 0) {
+        int xerrno = errno;
+
         pr_log_debug(DEBUG4, MOD_FACTS_VERSION ": error stat'ing '%s': %s",
-          path, strerror(errno));
+          path, strerror(xerrno));
+
+        errno = xerrno;
         return -1;
       }
 
@@ -437,7 +456,12 @@ static int facts_mlinfo_get(struct mlinfo *info, const char *path,
         }
 
       } else {
-        info->type = "file";
+        if (S_ISDIR(target_st.st_mode)) {
+          info->type = "dir";
+
+        } else {
+          info->type = "file";
+        }
       }
 
     } else {
@@ -851,6 +875,11 @@ MODRET facts_mff(cmd_rec *cmd) {
    * find the FIRST space in cmd->arg; the path is everything past that space.
    */
   ptr = strchr(cmd->arg, ' ');
+  if (ptr == NULL) {
+    pr_response_add_err(R_501, _("Invalid command syntax"));
+    return PR_ERROR(cmd);
+  }
+
   path = pstrdup(cmd->tmp_pool, ptr + 1);
 
   decoded_path = pr_fs_decode_path(cmd->tmp_pool, path);
@@ -1003,6 +1032,11 @@ MODRET facts_mfmt(cmd_rec *cmd) {
    * find the FIRST space in cmd->arg; the path is everything past that space.
    */
   ptr = strchr(cmd->arg, ' ');
+  if (ptr == NULL) {
+    pr_response_add_err(R_501, _("Invalid command syntax"));
+    return PR_ERROR(cmd);
+  }
+
   path = pstrdup(cmd->tmp_pool, ptr + 1);
 
   decoded_path = pr_fs_decode_path(cmd->tmp_pool, path);
@@ -1123,8 +1157,17 @@ MODRET facts_mlsd(cmd_rec *cmd) {
 
   /* Determine whether to display symlinks as such. */
   ptr = get_param_ptr(TOPLEVEL_CONF, "ShowSymlinks", FALSE);
-  if (ptr &&
-      *ptr == TRUE) {
+  if (ptr != NULL) {
+    if (*ptr == TRUE) {
+      flags |= FACTS_MLINFO_FL_SHOW_SYMLINKS;
+
+      if (facts_mlinfo_opts & FACTS_MLINFO_FL_SHOW_SYMLINKS_USE_SLINK) {
+        flags |= FACTS_MLINFO_FL_SHOW_SYMLINKS_USE_SLINK;
+      }
+    }
+
+  } else {
+    /* ShowSymlinks is documented as being 'on' by default. */
     flags |= FACTS_MLINFO_FL_SHOW_SYMLINKS;
 
     if (facts_mlinfo_opts & FACTS_MLINFO_FL_SHOW_SYMLINKS_USE_SLINK) {
@@ -1154,7 +1197,7 @@ MODRET facts_mlsd(cmd_rec *cmd) {
 
     } else {
       /* Handle the "DirFakeUser off" case (Bug#3715). */
-      fake_uid = session.uid;
+      fake_uid = (uid_t) -1;
     }
   }
 
@@ -1175,7 +1218,7 @@ MODRET facts_mlsd(cmd_rec *cmd) {
 
     } else {
       /* Handle the "DirFakeGroup off" case (Bug#3715). */
-      fake_gid = session.gid;
+      fake_gid = (gid_t) -1;
     }
   }
 
@@ -1189,6 +1232,8 @@ MODRET facts_mlsd(cmd_rec *cmd) {
       best_path, strerror(xerrno));
 
     pr_response_add_err(R_550, "%s: %s", path, strerror(xerrno));
+
+    errno = xerrno;
     return PR_ERROR(cmd);
   }
 
@@ -1209,6 +1254,10 @@ MODRET facts_mlsd(cmd_rec *cmd) {
     pr_signals_handle();
 
     rel_path = pdircat(cmd->tmp_pool, best_path, dent->d_name, NULL);
+    res = dir_check(cmd->tmp_pool, cmd, cmd->group, rel_path, &hidden);
+    if (!res || hidden) {
+      continue;
+    }
 
     /* Check that the file can be listed. */
     abs_path = dir_realpath(cmd->tmp_pool, rel_path);
@@ -1217,8 +1266,9 @@ MODRET facts_mlsd(cmd_rec *cmd) {
       
     } else {
       abs_path = dir_canonical_path(cmd->tmp_pool, rel_path);
-      if (abs_path == NULL)
+      if (abs_path == NULL) {
         abs_path = rel_path;
+      }
 
       res = dir_check_canon(cmd->tmp_pool, cmd, cmd->group, abs_path, &hidden);
     }
@@ -1264,6 +1314,15 @@ MODRET facts_mlsd(cmd_rec *cmd) {
 }
 
 MODRET facts_mlsd_cleanup(cmd_rec *cmd) {
+  const char *proto;
+
+  proto = pr_session_get_protocol(0);
+
+  /* Ignore this for SFTP connections. */
+  if (strncmp(proto, "sftp", 5) == 0) {
+    return PR_DECLINED(cmd);
+  }
+
   if (session.xfer.p) {
     destroy_pool(session.xfer.p);
   }
@@ -1500,6 +1559,27 @@ MODRET facts_opts_mlst(cmd_rec *cmd) {
   return PR_HANDLED(cmd);
 }
 
+MODRET facts_post_host(cmd_rec *cmd) {
+
+  /* If the HOST command changed the main_server pointer, reinitialize
+   * ourselves.
+   */
+  if (session.prev_server != NULL) {
+    int res;
+
+    facts_opts = 0;
+    facts_mlinfo_opts = 0;
+
+    res = facts_sess_init();
+    if (res < 0) {
+      pr_session_disconnect(&facts_module,
+        PR_SESS_DISCONNECT_SESSION_INIT_FAILED, NULL);
+    }
+  }
+
+  return PR_DECLINED(cmd);
+}
+
 /* Configuration handlers
  */
 
@@ -1576,8 +1656,15 @@ static int facts_sess_init(void) {
   }
 
   c = find_config(main_server->conf, CONF_PARAM, "FactsOptions", FALSE);
-  if (c) {
-    facts_mlinfo_opts = *((unsigned long *) c->argv[0]);
+  while (c != NULL) {
+    unsigned long opts;
+
+    pr_signals_handle();
+  
+    opts = *((unsigned long *) c->argv[0]);
+    facts_mlinfo_opts |= opts;
+
+    c = find_config_next(c, c->next, CONF_PARAM, "FactsOptions", FALSE);
   }
 
   facts_opts = FACTS_OPT_SHOW_MODIFY|FACTS_OPT_SHOW_PERM|FACTS_OPT_SHOW_SIZE|
@@ -1614,6 +1701,7 @@ static cmdtable facts_cmdtab[] = {
   { LOG_CMD_ERR,C_MLSD,		G_NONE,	facts_mlsd_cleanup, FALSE, FALSE },
   { CMD,	C_MLST,		G_DIRS,	facts_mlst, TRUE, FALSE, CL_DIRS },
   { CMD,	C_OPTS "_MLST", G_NONE, facts_opts_mlst, FALSE, FALSE },
+  { POST_CMD,	C_HOST,		G_NONE,	facts_post_host, FALSE, FALSE },
   { 0, NULL }
 };
 
